import uuid
from operator import itemgetter
from typing import Optional, List, Iterator, Dict, Any, Tuple, Iterable

import sqlalchemy as sa
from sqlalchemy.sql.elements import Null

from spinta import commands
from spinta.backends import Backend
from spinta.backends.constants import BackendOrigin
from spinta.components import Namespace, Base, Model, Property, Context, Config, EntryId, MetaData, Action
from spinta.core.enums import Access
from spinta.core.ufuncs import Expr
from spinta.datasets.components import Dataset, Resource, Param
from spinta.dimensions.comments.components import Comment
from spinta.dimensions.enum.components import Enums
from spinta.dimensions.lang.components import LangData
from spinta.dimensions.prefix.components import UriPrefix
from spinta.exceptions import InvalidIdType
from spinta.manifests.components import Manifest, ManifestSchema
from spinta.manifests.helpers import _load_manifest
from spinta.manifests.internal_sql.commands.auth import internal_authorized, get_namespace_highest_access
from spinta.manifests.internal_sql.components import InternalManifestRow, INTERNAL_MANIFEST_COLUMNS, \
    InternalManifestColumn, InternalSQLManifest
from spinta.manifests.tabular.components import ManifestRow, MANIFEST_COLUMNS
from spinta.manifests.tabular.helpers import ENUMS_ORDER_BY, sort, MODELS_ORDER_BY, DATASETS_ORDER_BY, \
    to_relative_model_name, PROPERTIES_ORDER_BY, _get_type_repr, _read_tabular_manifest_rows
from sqlalchemy_utils import UUIDType

from spinta.nodes import get_node
from spinta.spyna import unparse
from spinta.types.datatype import ArrayBackRef, Ref, Array, BackRef, Object
from spinta.types.text.components import Text
from spinta.utils.data import take
from spinta.utils.schema import NotAvailable, NA
from spinta.utils.types import is_str_uuid


def get_manifest(context: Context, manifest: InternalSQLManifest):
    if context.has('request.manifest'):
        return context.get('request.manifest')
    return manifest


def get_transaction_connection(context: Context):
    if context.has('transaction.manifest'):
        return context.get('transaction.manifest').connection
    return None


def can_return_namespace_data(context: Context, manifest: InternalSQLManifest, full_name: str, item, parents: list, action: Action):
    if full_name.startswith('_'):
        return False
    if not internal_authorized(
        context,
        full_name,
        get_namespace_highest_access(
            context,
            manifest,
            full_name
        ),
        action,
        parents
    ):
        return False

    return True


def select_full_table(table, extra_cols=None):
    if extra_cols is None:
        extra_cols = []

    return sa.select([
        table,
        sa.literal_column("prepare IS NULL").label("prepare_is_null")],
        *extra_cols
    )


def read_initial_schema(context: Context, manifest: InternalSQLManifest):
    conn = context.get('transaction.manifest').connection
    table = manifest.table
    stmt = select_full_table(table).where(table.c.path == None).order_by(table.c.index)
    rows = conn.execute(stmt)
    yield from internal_to_schema(manifest, rows)


def internal_to_schema(manifest: InternalSQLManifest, rows):
    converted = convert_sql_to_tabular_rows(list(rows))
    yield from _read_tabular_manifest_rows(path=manifest.path, rows=converted, allow_updates=True)


def read_schema(path: str):
    engine = sa.create_engine(path)
    with engine.connect() as conn:
        yield from _read_all_sql_manifest_rows(path, conn)


def get_object_from_id(context: Context, manifest: InternalSQLManifest, uid):
    conn = context.get('transaction.manifest').connection
    table = manifest.table
    results = conn.execute(select_full_table(table).where(
        table.c.id == uid
    ).limit(1))
    for item in results:
        return item
    return None


def update_schema_with_external(schema, external: dict):
    for id_, item in schema:
        if item['type'] == 'model':
            if external['dataset']:
                if '/' not in item['name']:
                    item['name'] = f'{external["dataset"]}/{item["name"]}'
                item['external']['dataset'] = external['dataset']
            if external['resource']:
                item['external']['resource'] = external['resource']
            if item['properties']:
                for prop in item['properties'].values():
                    if 'model' in prop and prop['model']:
                        if '/' not in prop['model'] and external['dataset']:
                            prop['model'] = f'{external["dataset"]}/{prop["model"]}'
            if item['base']:
                base = item['base']
                if base['parent']:
                    if '/' not in base['parent'] and external['dataset']:
                        base['parent'] = f'{external["dataset"]}/{base["parent"]}'
        yield id_, item


def _traverse_properties(prop: dict, ds: str):
    if prop['type'] not in ['backref', 'array_backref']:
        if 'model' in prop:
            model_name = prop['model']
            if '/' in model_name or not ds:
                yield model_name
            else:
                yield f'{ds}/{model_name}'
        elif 'items' in prop:
            yield from _traverse_properties(prop['items'], ds)
        elif 'properties' in prop and isinstance(prop['properties'], dict):
            for prop_ in prop['properties'].values():
                yield from _traverse_properties(prop_, ds)


def load_required_models(context: Context, manifest: InternalSQLManifest, schema, model_list: list):
    for id_, item in schema:
        if item['type'] == 'model':
            ds = item['external']['dataset'] if item['external'] and item['external']['dataset'] else None
            if item['properties']:
                for prop in item['properties'].values():
                    for model_prop in _traverse_properties(prop, ds):
                        if model_prop not in model_list:
                            model_list.append(model_prop)
                            commands.get_model(context, manifest, model_prop)
            if item['base']:
                if item['base']['parent'] not in model_list:
                    model_list.append(item['base']['parent'])
                    commands.get_model(context, manifest, item['base']['parent'])
        yield id_, item


def get_namespace_partial_data(
    context: Context,
    manifest: InternalSQLManifest,
    namespace: str,
    parents: list,
    action: Action,
    recursive: bool = False,
):
    conn = context.get('transaction.manifest').connection
    table = manifest.table
    parents = parents.copy()
    parents.append(namespace)

    results = conn.execute(select_full_table(table).where(
        sa.and_(
            table.c.mpath.startswith(namespace),
            table.c.mpath != namespace,
            sa.or_(
                table.c.dim == 'ns',
                table.c.dim == 'dataset',
                table.c.dim == 'model'
            )
        )
    ).order_by(table.c.mpath))
    result = []
    recursive_list = []
    for item in results:
        item = item._asdict()
        if item['path'] == namespace or item['mpath'] == namespace:
            continue

        type_ = 'ns'
        if item['dim'] == 'ns' or item['dim'] == 'dataset':
            name = item['name']
        else:
            type_ = 'model'
            name = item['path']

        name = name[len(namespace):]
        if name[0] == '/':
            name = name[1:]
        split = name.split('/')
        full_name = f'{namespace}/{split[0]}' if namespace else split[0]
        if len(split) == 1:
            result.append(split[0])
            if can_return_namespace_data(context, manifest, full_name, item, parents, action=action):
                if recursive and type_ == 'ns' and full_name not in recursive_list:
                    recursive_list.append(full_name)
                yield {
                    '_type': type_,
                    'name': f'{full_name}/:ns' if type_ == 'ns' else full_name,
                    'title': item['title'] or '',
                    'description': item['description'] or ''
                }
        elif split[0] not in result:
            result.append(split[0])
            if can_return_namespace_data(context, manifest, full_name, item, parents, action=action):
                if recursive and full_name not in recursive_list:
                    recursive_list.append(full_name)
                yield {
                    '_type': 'ns',
                    'name': f'{full_name}/:ns',
                    'title': '',
                    'description': ''
                }

    if recursive and recursive_list:
        for item in recursive_list:
            yield from get_namespace_partial_data(
                context,
                manifest,
                item,
                recursive=recursive,
                parents=parents,
                action=action
            )


def load_internal_manifest_nodes(
    context: Context,
    manifest: InternalSQLManifest,
    schemas: Iterable[ManifestSchema],
    *,
    link: bool = False,
    ignore_types: list = []
) -> None:
    to_link = []
    config = context.get('config')
    for eid, schema in schemas:
        if schema.get('type') == 'manifest':
            _load_manifest(context, manifest, schema, eid, reset=False)
        else:
            node = _load_internal_manifest_node(context, config, manifest, None, eid, schema)
            if node.type not in ignore_types:
                commands.set_node(context, manifest, node.type, node.name, node)
                if link:
                    to_link.append(node)

    if to_link:
        for node in to_link:
            commands.link(context, node)


def _load_internal_manifest_node(
    context: Context,
    config: Config,
    manifest: Manifest,
    source: Optional[Manifest],
    eid: EntryId,
    data: dict,
) -> MetaData:
    node = get_node(context, config, manifest, eid, data, check=False)
    node.eid = eid
    node.type = data['type']
    node.parent = manifest
    node.manifest = manifest
    commands.load(context, node, data, manifest, source=source)
    return node


def load_internal_namespace_from_name(
    context: Context,
    manifest: InternalSQLManifest,
    path: str,
    *,
    # Drop last element from path which is usually a model name.
    drop: bool = True,
) -> Namespace:
    ns: Optional[Namespace] = None
    parent: Optional[Namespace] = None
    parts: List[str] = []
    parts_ = [p for p in path.split('/') if p]
    if drop:
        parts_ = parts_[:-1]
    objects = manifest.get_objects()
    for part in parts_:
        parts.append(part)
        name = '/'.join(parts)

        ns = objects['ns'].get(name, None)
        if ns is None:
            ns = Namespace()
            data = {
                'type': 'ns',
                'name': name,
                'title': '',
                'description': '',
            }
            commands.load(context, ns, data, manifest)
            ns.generated = True

        if parent is not None:
            if ns.name == parent.name:
                raise RuntimeError(f"Self reference in {path!r}.")

            ns.parent = parent or manifest

            if part and part not in parent.names:
                parent.names[part] = ns
        else:
            ns.parent = manifest

        parent = ns
    return ns


def get_table_structure(meta: sa.MetaData):
    table = sa.Table(
        '_manifest',
        meta,
        sa.Column("index", sa.BIGINT),
        sa.Column("id", UUIDType, primary_key=True),
        sa.Column("parent", UUIDType),
        sa.Column("depth", sa.Integer),
        sa.Column("path", sa.String),
        sa.Column("mpath", sa.String),
        sa.Column("dim", sa.String),
        sa.Column("name", sa.String),
        sa.Column("type", sa.String),
        sa.Column("ref", sa.String),
        sa.Column("source", sa.String),
        sa.Column("prepare", sa.JSON),
        sa.Column("level", sa.Integer),
        sa.Column("access", sa.String),
        sa.Column("uri", sa.String),
        sa.Column("title", sa.String),
        sa.Column("description", sa.String),
    )
    return table


def _read_all_sql_manifest_rows(
    path: Optional[str],
    conn: sa.engine.Connection,
    *,
    rename_duplicates: bool = True
):
    meta = sa.MetaData(conn)
    inspector = sa.inspect(conn.engine)
    table_names = inspector.get_table_names()
    if '_manifest' in table_names:
        table = get_table_structure(meta)
        full_table = select_full_table(table)
        stmt = full_table.order_by(
            table.c.index
        )
        rows = conn.execute(stmt)
        converted = convert_sql_to_tabular_rows(list(rows))
        yield from _read_tabular_manifest_rows(path=path, rows=converted, rename_duplicates=rename_duplicates)


def write_internal_sql_manifest(context: Context, dsn: str, manifest: Manifest):
    engine = sa.create_engine(dsn)
    inspect = sa.inspect(engine)
    with engine.connect() as conn:
        trans = conn.begin()
        try:
            meta = sa.MetaData(conn)
            table = get_table_structure(meta)
            if inspect.has_table('_manifest'):
                conn.execute(table.delete())
            else:
                table.create()
            rows = datasets_to_sql(context, manifest)
            index = 0
            for row in rows:
                row['index'] = index
                index += 1
                conn.execute(table.insert().values(row))

            trans.commit()
        except Exception as e:
            trans.rollback()
            raise e


def _handle_id(item_id: Any):
    if item_id:
        if is_str_uuid(item_id):
            return uuid.UUID(item_id, version=4)
        elif isinstance(item_id, uuid.UUID):
            return item_id
        else:
            raise InvalidIdType(id=item_id, id_type=type(item_id))
    return uuid.uuid4()


def _compare_bases(base_0, base_1):
    if isinstance(base_0, Base) and isinstance(base_1, Base):
        converted_0 = list(_base_to_sql(base_0))[0]
        converted_0["id"] = base_0.id
        converted_1 = list(_base_to_sql(base_1))[0]
        converted_1["id"] = base_1.id
        return converted_0 == converted_1
    return base_0 == base_1


def datasets_to_sql(
    context: Context,
    manifest: Manifest,
    *,
    external: bool = True,  # clean content of source and prepare
    access: Access = Access.private,
    internal: bool = False,  # internal models with _ prefix like _txn
    order_by: InternalManifestColumn = None,
) -> Iterator[InternalManifestRow]:
    yield from _prefixes_to_sql(manifest.prefixes)
    yield from _backends_to_sql(manifest.backends)
    yield from _namespaces_to_sql(commands.get_namespaces(context, manifest))
    yield from _enums_to_sql(
        manifest.enums,
        external=external,
        access=access,
        order_by=order_by,
    )

    seen_datasets = set()
    dataset = {
        "id": None,
        "path": None,
        "mpath": None,
        "item": None,
        "depth": 0
    }
    resource = {
        "id": None,
        "path": None,
        "mpath": None,
        "item": None,
        "depth": 0
    }
    base = {
        "id": None,
        "path": None,
        "mpath": None,
        "item": None,
        "depth": 0
    }
    models = commands.get_models(context, manifest)
    models = models if internal else take(models)
    models = sort(MODELS_ORDER_BY, models.values(), order_by)

    for model in models:
        if model.access < access:
            continue

        if model.external:
            if dataset["item"] is None or (model.external.dataset and dataset["item"].name != model.external.dataset.name):
                dataset["item"] = model.external.dataset
                if dataset["item"]:
                    seen_datasets.add(dataset["item"].name)
                    resource["item"] = None
                    for item in _dataset_to_sql(
                        dataset["item"],
                        external=external,
                        access=access,
                        order_by=order_by,
                    ):
                        yield item
                        if item["dim"] == "dataset":
                            dataset["id"] = item["id"]
                            dataset["path"] = item["path"]
                            dataset["mpath"] = item["mpath"]
                            dataset["depth"] = item["depth"]
            if not model.external.dataset:
                dataset["item"] = None
                resource["item"] = None
                base["item"] = None

            if external and model.external.resource and (
                resource["item"] is None or
                resource["item"].name != model.external.resource.name
            ):
                resource["item"] = model.external.resource
                if resource["item"]:
                    parent_id = None
                    depth = 0
                    path = ''
                    mpath = ''
                    if dataset["item"]:
                        parent_id = dataset["id"]
                        depth = dataset["depth"] + 1
                        path = dataset["path"]
                        mpath = dataset["mpath"]
                    for item in _resource_to_sql(
                        resource["item"],
                        external=external,
                        access=access,
                        parent_id=parent_id,
                        path=path,
                        mpath=mpath,
                        depth=depth,
                    ):
                        yield item
                        if item["dim"] == "resource":
                            resource["id"] = item["id"]
                            resource["path"] = item["path"]
                            resource["mpath"] = item["mpath"]
                            resource["depth"] = item["depth"]
            elif external:
                if resource["item"] is not None and not model.external.resource:
                    resource["item"] = None
                    base["item"] = None

        if model.base and (not base["item"] or not _compare_bases(model.base, base["item"])):
            base["item"] = model.base
            parent_id = None
            depth = 0
            path = ''
            mpath = ''
            if resource["item"]:
                parent_id = resource["id"]
                depth = resource["depth"] + 1
                path = resource["path"]
                mpath = resource["mpath"]
            elif dataset["item"]:
                parent_id = dataset["id"]
                depth = dataset["depth"] + 1
                path = dataset["path"]
                mpath = dataset["mpath"]
            for item in _base_to_sql(
                base=base["item"],
                parent_id=parent_id,
                depth=depth,
                path=path,
                mpath=mpath,
            ):
                yield item
                if item["dim"] == "base":
                    base["id"] = item["id"]
                    base["path"] = item["path"]
                    base["mpath"] = item["mpath"]
                    base["depth"] = item["depth"]
        elif base["item"] and not model.base:
            base["item"] = None
        parent_id = None
        depth = 0
        path = dataset["path"] if dataset["item"] else ""
        mpath = ''
        if base["item"]:
            parent_id = base["id"]
            depth = base["depth"] + 1
            mpath = base["mpath"]
        elif resource["item"]:
            parent_id = resource["id"]
            depth = resource["depth"] + 1
            mpath = resource["mpath"]
        elif dataset["item"]:
            parent_id = dataset["id"]
            depth = dataset["depth"] + 1
            mpath = dataset["mpath"]
        yield from _model_to_sql(
            model,
            external=external,
            access=access,
            order_by=order_by,
            parent_id=parent_id,
            depth=depth,
            path=path,
            mpath=mpath
        )

    datasets = sort(DATASETS_ORDER_BY, commands.get_datasets(context, manifest).values(), order_by)
    for dataset in datasets:
        if dataset.name in seen_datasets:
            continue
        parent_id = None
        depth = 0
        path = ''
        mpath = ''
        for item in _dataset_to_sql(
            dataset,
            external=external,
            access=access,
            order_by=order_by,
        ):
            yield item
            if item["dim"] == "dataset":
                parent_id = item["id"]
                depth = int(item["depth"]) + 1
                path = item["path"]
                mpath = item["mpath"]

        for resource in dataset.resources.values():
            yield from _resource_to_sql(
                resource,
                parent_id=parent_id,
                depth=depth,
                path=path,
                mpath=mpath
            )


def _prefixes_to_sql(
    prefixes: Dict[str, UriPrefix],
    parent_id: uuid.UUID = None,
    depth: int = 0,
    path: str = None,
    mpath: str = None
) -> Iterator[InternalManifestRow]:
    for name, prefix in prefixes.items():
        item_id = _handle_id(prefix.id)
        yield to_row(INTERNAL_MANIFEST_COLUMNS, {
            'id': item_id,
            'parent': parent_id,
            'depth': depth,
            'path': path,
            'mpath': '/'.join([mpath, name] if mpath else [name]),
            'dim': 'prefix',
            'name': name,
            'type': prefix.type,
            'ref': name,
            'uri': prefix.uri,
            'title': prefix.title,
            'description': prefix.description,
            'prepare': _handle_prepare(NA)
        })


def _namespaces_to_sql(
    namespaces: Dict[str, Namespace],
    parent_id: uuid.UUID = None,
    depth: int = 0,
    path: str = None,
    mpath: str = None
) -> Iterator[InternalManifestRow]:
    namespaces = {
        k: ns
        for k, ns in namespaces.items() if not ns.generated
    }

    for name, ns in namespaces.items():
        item_id = _handle_id(ns.id)
        yield to_row(INTERNAL_MANIFEST_COLUMNS, {
            'id': item_id,
            'parent': parent_id,
            'depth': depth,
            'path': path,
            'mpath': '/'.join([mpath, name] if mpath else [name]),
            'dim': 'ns',
            'name': name,
            'type': ns.type,
            'ref': name,
            'title': ns.title,
            'description': ns.description,
            'prepare': _handle_prepare(NA)
        })


def _enums_to_sql(
    enums: Optional[Enums],
    parent_id: uuid.UUID = None,
    depth: int = 0,
    path: str = None,
    mpath: str = None,
    external: bool = True,
    access: Access = Access.private,
    order_by: InternalManifestColumn = None,
) -> Iterator[InternalManifestRow]:
    if enums is None:
        return
    for name, enum in enums.items():
        items = sort(ENUMS_ORDER_BY, enum.values(), order_by)
        new_parent_id = _handle_id("")
        mpath_name = name if name else str(new_parent_id)
        new_mpath = '/'.join([mpath, mpath_name] if mpath else [mpath_name])
        yield to_row(INTERNAL_MANIFEST_COLUMNS, {
            'id': new_parent_id,
            'parent': parent_id,
            'depth': depth,
            'path': path,
            'mpath': new_mpath,
            'dim': 'enum',
            'name': name,
            'type': 'enum',
            'ref': name,
            'prepare': _handle_prepare(NA),
        })
        for item in items:
            if item.access is not None and item.access < access:
                continue
            new_item_id = _handle_id(item.id)
            new_item_mpath = '/'.join([new_mpath, str(new_item_id)] if new_mpath else [str(new_item_id)])
            yield to_row(INTERNAL_MANIFEST_COLUMNS, {
                'id': new_item_id,
                'parent': new_parent_id,
                'depth': depth + 1,
                'path': path,
                'mpath': new_item_mpath,
                'dim': 'enum.item',
                'source': item.source if external else None,
                'prepare': _handle_prepare(item.prepare),
                'access': item.given.access,
                'title': item.title,
                'description': item.description,
                'level': item.level,
            })
            yield from _lang_to_sql(item.lang, path=path, mpath=new_item_mpath, depth=depth + 2, parent_id=new_item_id)


def _lang_to_sql(
    lang: Optional[LangData],
    parent_id: uuid.UUID = None,
    depth: int = 0,
    path: str = None,
    mpath: str = None,
) -> Iterator[InternalManifestRow]:
    if lang is None:
        return
    for name, data in sorted(lang.items(), key=itemgetter(0)):
        item_id = _handle_id(data.get("id"))
        yield to_row(INTERNAL_MANIFEST_COLUMNS, {
            'id': item_id,
            'parent': parent_id,
            'depth': depth,
            'path': path,
            'mpath': '/'.join([mpath, name] if mpath else [name]),
            'dim': 'lang',
            'name': name,
            'type': 'lang',
            'ref': name,
            'title': data['title'],
            'description': data['description'],
            'prepare': _handle_prepare(NA)
        })


def _comments_to_sql(
    comments: Optional[List[Comment]],
    access: Access = Access.private,
    parent_id: uuid.UUID = None,
    depth: int = 0,
    path: str = None,
    mpath: str = None
) -> Iterator[InternalManifestRow]:
    if comments is None:
        return
    for comment in comments:
        if comment.access < access:
            return
        new_id = _handle_id(comment.id)
        yield to_row(INTERNAL_MANIFEST_COLUMNS, {
            'id': new_id,
            'parent': parent_id,
            'depth': depth,
            'path': path,
            'mpath': '/'.join([mpath, str(new_id)] if mpath else [str(new_id)]),
            'name': comment.parent if comment.parent else None,
            'dim': 'comment',
            'type': 'comment',
            'ref': comment.parent,
            'source': comment.author,
            'access': comment.given.access,
            'title': comment.created,
            'description': comment.comment,
            'prepare': _handle_prepare(NA)
        })


def _backends_to_sql(
    backends: Dict[str, Backend],
    parent_id: uuid.UUID = None,
    depth: int = 0,
    path: str = None,
    mpath: str = None
) -> Iterator[InternalManifestRow]:
    for name, backend in backends.items():
        new_id = _handle_id(backend.config.get("id"))
        yield to_row(INTERNAL_MANIFEST_COLUMNS, {
            'id': new_id,
            'parent': parent_id,
            'depth': depth,
            'path': path,
            'mpath': '/'.join([mpath, name] if mpath else [name]),
            'dim': 'resource',
            'name': name,
            'type': backend.type,
            'source': backend.config.get('dsn'),
            'prepare': _handle_prepare(NA)
        })


def _dataset_to_sql(
    dataset: Dataset,
    parent_id: uuid.UUID = None,
    path: str = None,
    mpath: str = None,
    depth: int = 0,
    external: bool = True,
    access: Access = Access.private,
    order_by: InternalManifestColumn = None,
) -> Iterator[InternalManifestRow]:
    dataset_id = _handle_id(dataset.id)
    new_path = '/'.join([path, dataset.name] if path else [dataset.name])
    new_mpath = '/'.join([mpath, dataset.name] if mpath else [dataset.name])
    yield to_row(INTERNAL_MANIFEST_COLUMNS, {
        'id': dataset_id,
        'parent': parent_id,
        'depth': depth,
        'path': new_path,
        'mpath': new_mpath,
        'dim': 'dataset',
        'name': dataset.name,
        'dataset': dataset.name,
        'level': dataset.level,
        'access': dataset.given.access,
        'title': dataset.title,
        'description': dataset.description,
        'prepare': _handle_prepare(NA)
    })
    yield from _lang_to_sql(dataset.lang, parent_id=dataset_id, depth=depth + 1, path=new_path, mpath=new_mpath)
    yield from _prefixes_to_sql(dataset.prefixes, parent_id=dataset_id, depth=depth + 1, path=new_path, mpath=new_mpath)
    yield from _enums_to_sql(
        dataset.ns.enums,
        external=external,
        access=access,
        order_by=order_by,
        parent_id=dataset_id,
        depth=depth + 1,
        path=new_path,
        mpath=new_mpath
    )


def _params_to_sql(
    params: List[Param],
    parent_id: uuid.UUID = None,
    depth: int = 0,
    path: str = None,
    mpath: str = None
) -> Iterator[InternalManifestRow]:
    if not params:
        return
    for param in params:
        param_base_id = _handle_id("")
        new_mpath = '/'.join([mpath, param.name] if mpath else [param.name])
        for i, (source, prepare) in enumerate(zip(param.source, param.prepare)):
            new_id = _handle_id('')
            if param.id is not None:
                new_id = _handle_id(param.id)
            prepare = _handle_prepare(prepare)
            if i == 0:
                yield to_row(INTERNAL_MANIFEST_COLUMNS, {
                    'id': param_base_id,
                    'parent': parent_id,
                    'depth': depth,
                    'path': path,
                    'mpath': new_mpath,
                    'dim': 'param',
                    'name': param.name,
                    'type': 'param',
                    'ref': param.name,
                    'source': source,
                    'title': param.title,
                    'description': param.description
                })
            yield to_row(INTERNAL_MANIFEST_COLUMNS, {
                'id': new_id,
                'parent': param_base_id,
                'depth': depth + 1,
                'path': path,
                'mpath': '/'.join([new_mpath, str(new_id)] if new_mpath else [str(new_id)]),
                'dim': 'param.item',
                'source': source,
                'prepare': prepare
            })


def _resource_to_sql(
    resource: Resource,
    parent_id: uuid.UUID = None,
    depth: int = 0,
    path: str = None,
    mpath: str = None,
    external: bool = True,
    access: Access = Access.private,
) -> Iterator[InternalManifestRow]:
    backend = resource.backend
    new_mpath = '/'.join([mpath, resource.name] if mpath else [resource.name])
    item_id = _handle_id(resource.id)
    yield to_row(INTERNAL_MANIFEST_COLUMNS, {
        'id': item_id,
        'parent': parent_id,
        'depth': depth,
        'path': path,
        'mpath': new_mpath,
        'dim': 'resource',
        'name': resource.name,
        'source': resource.external if external else None,
        'prepare': _handle_prepare(resource.prepare),
        'type': resource.type,
        'ref': (
            backend.name
            if (
                external and
                backend and
                backend.origin != BackendOrigin.resource
            )
            else None
        ),
        'level': resource.level,
        'access': resource.given.access,
        'title': resource.title,
        'description': resource.description,
    })
    yield from _params_to_sql(resource.params, parent_id=item_id, depth=depth + 1, path=path, mpath=new_mpath)
    yield from _comments_to_sql(resource.comments, access=access, parent_id=item_id, depth=depth + 1, path=path,
                                mpath=new_mpath)
    yield from _lang_to_sql(resource.lang, parent_id=item_id, depth=depth + 1, path=path, mpath=new_mpath)


def _base_to_sql(
    base: Base,
    parent_id: uuid.UUID = None,
    depth: int = 0,
    path: str = None,
    mpath: str = None
) -> Iterator[InternalManifestRow]:
    item_id = _handle_id(base.id)
    new_path = '/'.join([path, base.name] if path else [base.name])
    new_mpath = '/'.join([mpath, base.name] if mpath else [base.name])
    data = {
        'id': item_id,
        'parent': parent_id,
        'depth': depth,
        'path': new_path,
        'mpath': new_mpath,
        'dim': 'base',
        'name': base.name,
        'level': base.level.value if base.level else None,
        'prepare': _handle_prepare(NA)
    }
    if base.pk:
        data['ref'] = ', '.join([pk.place for pk in base.pk])
    yield to_row(INTERNAL_MANIFEST_COLUMNS, data)
    yield from _lang_to_sql(base.lang, parent_id=item_id, depth=depth + 1, path=new_path, mpath=new_mpath)


def _model_to_sql(
    model: Model,
    external: bool = True,
    access: Access = Access.private,
    order_by: InternalManifestColumn = None,
    parent_id: uuid.UUID = None,
    depth: int = 0,
    path: str = None,
    mpath: str = None
) -> Iterator[InternalManifestRow]:
    item_id = _handle_id(model.id)
    name = model.name
    if model.external and model.external.dataset:
        name = to_relative_model_name(
            model,
            model.external.dataset,
        )
    new_mpath = '/'.join([mpath, name] if mpath else [name])
    new_path = '/'.join([path, name] if path else [name])
    data = {
        'id': item_id,
        'parent': parent_id,
        'depth': depth,
        'path': new_path,
        'mpath': new_mpath,
        'dim': 'model',
        'name': name,
        'level': model.level.value if model.level else None,
        'access': model.given.access,
        'title': model.title,
        'description': model.description,
        'uri': model.uri if model.uri else None,
    }

    if external and model.external:
        data.update({
            'source': model.external.name,
            'prepare': _handle_prepare(model.external.prepare),
        })
        if (
            not model.external.unknown_primary_key and
            all(p.access >= access for p in model.external.pkeys)
        ):
            # Add `ref` only if all properties are available in the
            # resulting manifest.
            data['ref'] = ', '.join([
                p.name for p in model.external.pkeys
            ])

    hide_list = []
    if model.external:
        if not model.external.unknown_primary_key:
            hide_list = [model.external.pkeys]
    yield to_row(INTERNAL_MANIFEST_COLUMNS, data)
    yield from _params_to_sql(model.params, parent_id=item_id, depth=depth + 1, path=new_path, mpath=new_mpath)
    yield from _comments_to_sql(model.comments, access=access, parent_id=item_id, depth=depth + 1, path=new_path,
                                mpath=new_mpath)
    yield from _lang_to_sql(model.lang, parent_id=item_id, depth=depth + 1, path=new_path, mpath=new_mpath)
    yield from _unique_to_sql(model.unique, hide_list, parent_id=item_id, depth=depth + 1, path=new_path,
                              mpath=new_mpath)

    props = sort(PROPERTIES_ORDER_BY, model.properties.values(), order_by)
    for prop in props:
        yield from _property_to_sql(
            prop,
            external=external,
            access=access,
            order_by=order_by,
            parent_id=item_id,
            depth=depth + 1,
            path=new_path,
            mpath=new_mpath
        )


def _unique_to_sql(
    model_unique_data,
    hide_list: List,
    parent_id: uuid.UUID = None,
    depth: int = 0,
    path: str = None,
    mpath: str = None
) -> Iterator[InternalManifestRow]:
    if not model_unique_data:
        return
    for row in model_unique_data:
        if row not in hide_list:
            item_id = _handle_id("")
            yield to_row(INTERNAL_MANIFEST_COLUMNS, {
                'id': item_id,
                'parent': parent_id,
                'depth': depth,
                'path': path,
                'mpath': '/'.join([mpath, str(item_id)] if mpath else [item_id]),
                'dim': 'unique',
                'type': 'unique',
                'ref': ', '.join([r.name for r in row]),
                'prepare': _handle_prepare(NA)
            })


def _property_to_sql(
    prop: Property,
    external: bool = True,
    access: Access = Access.private,
    order_by: InternalManifestColumn = None,
    parent_id: uuid.UUID = None,
    depth: int = 0,
    path: str = None,
    mpath: str = None
) -> Iterator[InternalManifestRow]:
    if prop.name.startswith('_'):
        return

    if prop.access and prop.access < access:
        return

    item_id = _handle_id(prop.id)
    new_path = '/'.join([path, prop.name] if path else [prop.name])
    new_mpath = '/'.join([mpath, prop.name] if mpath else [prop.name])
    type_ = _get_type_repr(prop.dtype)
    data = {
        'id': item_id,
        'parent': parent_id,
        'depth': depth,
        'path': new_path,
        'mpath': new_mpath,
        'dim': 'property',
        'name': prop.given.name,
        'type': type_,
        'level': prop.level.value if prop.level else None,
        'access': prop.given.access,
        'uri': prop.uri,
        'title': prop.title,
        'description': prop.description,
    }

    if external and prop.external:
        if isinstance(prop.external, list):
            # data['source'] = ', '.join(x.name for x in prop.external)
            # data['prepare'] = ', '.join(
            #     unparse(x.prepare or NA)
            #     for x in prop.external if x.prepare
            # )
            raise DeprecationWarning(
                "Source can't be a list, use prepare instead."
            )
        elif prop.external:
            data['source'] = prop.external.name
            data['prepare'] = _handle_prepare(prop.external.prepare)

    yield_rows = []

    if isinstance(prop.dtype, (Array, ArrayBackRef)):
        yield_array_row = prop.dtype.items
        yield_rows.append(yield_array_row)

    elif isinstance(prop.dtype, (Ref, BackRef)):
        model = prop.model
        data['ref'] = prop.dtype.model.name

        if model.external and model.external.dataset:
            if isinstance(prop.dtype, Ref):
                pkeys = prop.dtype.model.external.pkeys
                rkeys = prop.dtype.refprops

                if rkeys and pkeys != rkeys:
                    rkeys = ', '.join([p.place for p in rkeys])
                    data['ref'] += f'[{rkeys}]'
            else:
                rkey = prop.dtype.refprop.place
                if prop.dtype.explicit:
                    data['ref'] += f'[{rkey}]'

        if prop.dtype.properties:
            for obj_prop in prop.dtype.properties.values():
                yield_rows.append(obj_prop)

    elif isinstance(prop.dtype, Object):
        for obj_prop in prop.dtype.properties.values():
            yield_rows.append(obj_prop)

    elif isinstance(prop.dtype, Text):
        for lang_prop in prop.dtype.langs.values():
            yield_rows.append(lang_prop)

    elif prop.enum is not None:
        data['ref'] = prop.given.enum
    elif prop.unit is not None:
        data['ref'] = prop.given.unit

    if data['name']:
        yield to_row(INTERNAL_MANIFEST_COLUMNS, data)
    yield from _comments_to_sql(prop.comments, access=access, parent_id=item_id, depth=depth + 1, path=new_path,
                                mpath=new_mpath)
    yield from _lang_to_sql(prop.lang, parent_id=item_id, depth=depth + 1, path=new_path, mpath=new_mpath)
    yield from _enums_to_sql(
        prop.enums,
        external=external,
        access=access,
        order_by=order_by,
        parent_id=item_id,
        depth=depth + 1,
        path=new_path,
        mpath=new_mpath
    )
    if yield_rows:
        for yield_row in yield_rows:
            if not yield_row:
                continue

            yield from _property_to_sql(yield_row, external=external, access=access, order_by=order_by, parent_id=item_id, depth=depth + 1, path=new_path, mpath=new_mpath)


def _value_or_null(value: Any):
    if isinstance(value, Null) or value or value is False or value == 0:
        return value
    return None


def _value_or_empty(value: Any):
    if isinstance(value, Null) or value or value is False or value == 0:
        return value
    return ''


def to_row(keys, values) -> InternalManifestRow:
    return {k: _value_or_null(values.get(k)) for k in keys}


def to_row_tabular(keys, values) -> ManifestRow:
    result = {}
    for key in keys:
        if key == "prepare" and values.get("prepare_is_null") is not None and not values.get("prepare_is_null"):
            result[key] = unparse(values.get("prepare"))
        else:
            result[key] = _value_or_empty(values.get(key))
    return result


def _handle_prepare(prepare: Any):
    if isinstance(prepare, NotAvailable):
        prepare = sa.null()
    elif isinstance(prepare, Expr):
        prepare = prepare.todict()
    return prepare


def convert_sql_to_tabular_rows(rows: list) -> Iterator[Tuple[str, List[str]]]:
    previous_row = {}
    data_row = {}
    enum_data = {}
    param_data = {}
    meta_dimensions = {
        "current": None,
        "previous": None,
        'dataset': None,
        'resource': None,
        'base': None
    }
    yield 1, MANIFEST_COLUMNS
    line = 2
    for i, row in enumerate(rows):
        row = dict(row)
        _update_meta_dimensions(meta_dimensions, row)
        is_first = row["dim"] != previous_row.get("dim")
        base, resource, dataset = _requires_end_marker(row, meta_dimensions)
        if dataset:
            yield line, list(to_row_tabular(MANIFEST_COLUMNS, {'dataset': '/'}).values())
            line += 1
        elif resource:
            yield line, list(to_row_tabular(MANIFEST_COLUMNS, {'resource': '/'}).values())
            line += 1
        elif base:
            yield line, list(to_row_tabular(MANIFEST_COLUMNS, {'base': '/'}).values())
            line += 1
        if _requires_seperator(row, previous_row, meta_dimensions):
            yield line, list(to_row_tabular(MANIFEST_COLUMNS, {}).values())
            line += 1
        dimension = row["dim"]
        if dimension == "prefix":
            data_row = _convert_prefixes(row, is_first)
        elif dimension == "ns":
            data_row = _convert_namespaces(row, is_first)
        elif dimension == "lang":
            data_row = _convert_lang(row, is_first)
        elif dimension == "comment":
            data_row = _convert_comment(row, is_first)
        elif dimension == "resource":
            data_row = _convert_resource(row)
        elif dimension == "enum":
            enum_data = _convert_enum(row)
        elif dimension == "enum.item":
            data_row = _convert_enum_item(row, enum_data, is_first)
        elif dimension == "unique":
            data_row = _convert_unique(row)
        elif dimension == "param":
            param_data = row
        elif dimension == "param.item":
            data_row = _convert_param(row, param_data, is_first)
        elif dimension == "dataset":
            data_row = _convert_dataset(row)
        elif dimension == "base":
            data_row = _convert_base(row)
        elif dimension == "property":
            data_row = _convert_property(row)
        elif dimension == "model":
            data_row = _convert_model(row)
        previous_row = row
        if dimension != "enum" and dimension != "param":
            yield line, list(data_row.values())
            line += 1


def _update_meta_dimensions(meta_dimensions: dict, row: InternalManifestRow):
    if row["dim"] == "dataset":
        meta_dimensions["dataset"] = row["depth"]
        meta_dimensions["resource"] = None
        meta_dimensions["base"] = None
    elif row["dim"] == "resource":
        meta_dimensions["resource"] = row["depth"]
        meta_dimensions["base"] = None
    elif row["dim"] == "base":
        meta_dimensions["base"] = row["depth"]
    if meta_dimensions["current"] != row["dim"] and row["dim"] in [
        "dataset", "resource", "base", "model", "property"
    ]:
        meta_dimensions["previous"] = meta_dimensions["current"]
        meta_dimensions["current"] = row["dim"]


def _requires_end_marker(row: InternalManifestRow, meta_dimensions: dict):
    base_end_marker = False
    resource_end_marker = False
    dataset_end_marker = False
    if row["dim"] not in ["dataset", "resource", "base", "property", "enum.item", "param.item"]:
        depth = row["depth"]
        if meta_dimensions["base"] is not None:
            if depth <= meta_dimensions["base"]:
                meta_dimensions["base"] = None
                base_end_marker = True
        if meta_dimensions["resource"]:
            if depth <= meta_dimensions["resource"] is not None:
                meta_dimensions["resource"] = None
                resource_end_marker = True
        if meta_dimensions["dataset"] is not None:
            if depth <= meta_dimensions["dataset"]:
                meta_dimensions["dataset"] = None
                dataset_end_marker = True
    return base_end_marker, resource_end_marker, dataset_end_marker


def _requires_seperator(row: InternalManifestRow, previous_row: InternalManifestRow, meta_dimensions: dict):
    primary_list = (None, "dataset", "resource", "base", "model", "property")

    def is_primary_dimension(dimension: str):
        if dimension in primary_list:
            return primary_list.index(dimension)
        return -1

    if row["dim"] == previous_row.get("dim"):
        return False

    previous_dim = is_primary_dimension(previous_row.get("dim"))
    current_dim = is_primary_dimension(row["dim"])
    if previous_row.get("dim") == "base":
        return False
    elif row["dim"] == "resource":
        return False
    elif current_dim > previous_dim != -1:
        return False
    elif current_dim == -1 and previous_dim != -1:
        return False
    elif current_dim == -1 and previous_dim == -1:
        if previous_row.get("dim") == "lang" or row.get("dim") == "lang":
            return False
        if (previous_row.get("dim") == "enum" and row["dim"] == "enum.item") or (previous_row.get("dim") == "param" and row["dim"] == "param.item"):
            return False
    elif current_dim != -1 and primary_list[current_dim] == "property" and previous_dim == -1:
        return False
    elif previous_dim == -1 and meta_dimensions["previous"] == meta_dimensions["current"]:
        return False
    return True


def _convert_model(row: InternalManifestRow):
    new = to_row_tabular(MANIFEST_COLUMNS, row)
    new["model"] = row["name"]
    return new


def _convert_property(row: InternalManifestRow):
    new = to_row_tabular(MANIFEST_COLUMNS, row)
    new["property"] = row["name"]
    return new


def _convert_base(row: InternalManifestRow):
    new = to_row_tabular(MANIFEST_COLUMNS, row)
    new["base"] = row["name"]
    return new


def _convert_dataset(row: InternalManifestRow):
    new = to_row_tabular(MANIFEST_COLUMNS, row)
    new["dataset"] = row["name"]
    return new


def _convert_param(row: InternalManifestRow, param_data: InternalManifestRow, first: bool = False):
    new = to_row_tabular(MANIFEST_COLUMNS, row)
    if not first:
        new["type"] = ''
        new["ref"] = ''
        new["title"] = ''
        new["description"] = ''
    else:
        new["type"] = 'param'
        new["ref"] = _value_or_empty(param_data["ref"])
        new["title"] = _value_or_empty(param_data["title"])
        new["description"] = _value_or_empty(param_data["description"])
    return new


def _convert_unique(row: InternalManifestRow):
    return to_row_tabular(MANIFEST_COLUMNS, row)


def _convert_enum_item(row: InternalManifestRow, enum_data: InternalManifestRow, first: bool = False):
    new = to_row_tabular(MANIFEST_COLUMNS, row)
    if not first:
        new["type"] = ''
        new["ref"] = ''
    else:
        new["type"] = "enum"
        new["ref"] = enum_data["ref"]
    return new


def _convert_resource(row: InternalManifestRow):
    new = to_row_tabular(MANIFEST_COLUMNS, row)
    new["resource"] = row["name"]
    return new


def _convert_comment(row: InternalManifestRow, first: bool = False):
    new = to_row_tabular(MANIFEST_COLUMNS, row)
    if not first:
        new["type"] = ''
    return new


def _convert_lang(row: InternalManifestRow, first: bool = False):
    new = to_row_tabular(MANIFEST_COLUMNS, row)
    if not first:
        new["type"] = ''
        new["ref"] = ''
    return new


def _convert_namespaces(row: InternalManifestRow, first: bool = False):
    new = to_row_tabular(MANIFEST_COLUMNS, row)
    if not first:
        new["type"] = ''
    return new


def _convert_prefixes(row: InternalManifestRow, first: bool = False):
    new = to_row_tabular(MANIFEST_COLUMNS, row)
    if not first:
        new["type"] = ''
    return new


def _convert_enum(row: InternalManifestRow):
    new = to_row_tabular(MANIFEST_COLUMNS, row)
    new["ref"] = _value_or_empty(new["ref"])
    return new



import importlib.metadata

from setuptools_scm import get_version as get_scm_version

from spinta.commands import get_version
from spinta.components import Context


@get_version.register(Context)
def get_version(context: Context):
    version = importlib.metadata.version('spinta')
    if version is None:
        version = get_scm_version()

    return {
        'implementation': {
            'name': 'Spinta',
            'version': version,
        },
        'uapi': {
            # Supported UAPI version:
            # https://ivpk.github.io/uapi/
            'version': '0.1.0',
        },
        'dsa': {
            # Supported DSA version:
            # https://ivpk.github.io/dsa/
            'version': '0.1.0',
        }
    }



import ujson as json

from spinta.formats.components import Format


class JsonLines(Format):
    content_type = 'application/x-json-stream'
    accept_types = {
        'application/x-json-stream',
    }
    params = {}

    def __call__(self, data):
        if isinstance(data, dict):
            data = iter([data])
        memory = next(data, None)

        for row in data:
            yield json.dumps({k: v for k, v in memory.items() if k != '_page'}, ensure_ascii=False) + '\n'
            memory = row

        if memory is not None:
            last_row = {}
            for k, v in memory.items():
                if k == '_page':
                    last_row[k] = {
                        'next': v
                    }
                else:
                    last_row[k] = v
            yield json.dumps(last_row, ensure_ascii=False) + '\n'



import sqlalchemy as sa

from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers.migrate.migrate import MigratePostgresMeta
from spinta.components import Context, Property
from spinta.utils.schema import NotAvailable


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, list, Property)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: list, new: Property, **kwargs):
    commands.migrate(context, backend, meta, table, old, new.dtype, **kwargs)


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, sa.Column, Property)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: sa.Column, new: Property, **kwargs):
    commands.migrate(context, backend, meta, table, old, new.dtype, **kwargs)


@commands.migrate.register(Context, PostgreSQL, MigratePostgresMeta, sa.Table, NotAvailable, Property)
def migrate(context: Context, backend: PostgreSQL, meta: MigratePostgresMeta, table: sa.Table,
            old: NotAvailable, new: Property, **kwargs):
    commands.migrate(context, backend, meta, table, old, new.dtype, **kwargs)



import dataclasses
from collections import defaultdict
from typing import Any, List, Union, Dict, Tuple, Callable

import sqlalchemy as sa
import geoalchemy2.types
from spinta.datasets.inspect.helpers import zipitems
from spinta.utils.itertools import ensure_list
from sqlalchemy.dialects.postgresql import JSONB, BIGINT, ARRAY, JSON
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.dialects import postgresql

import spinta.backends.postgresql.helpers.migrate.actions as ma
from spinta import commands
from spinta.backends.constants import TableType, BackendFeatures
from spinta.backends.helpers import get_table_name
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_pg_name, get_column_name
from spinta.backends.postgresql.helpers.migrate.actions import MigrationHandler
from spinta.backends.postgresql.helpers.name import name_changed, get_pg_constraint_name, get_pg_index_name, \
    nested_column_rename, get_pg_table_name, get_pg_column_name
from spinta.cli.helpers.migrate import MigrateRename
from spinta.components import Context, Model, Property
from spinta.exceptions import MigrateScalarToRefTooManyKeys
from spinta.types.datatype import Ref, File, Array, Object
from spinta.types.text.components import Text
from spinta.utils.nestedstruct import get_root_attr
from spinta.utils.schema import NA


def drop_all_indexes_and_constraints(inspector: Inspector, table: str, new_table: str, handler: MigrationHandler):
    constraints = inspector.get_unique_constraints(table)
    removed = []
    foreign_keys = inspector.get_foreign_keys(table)
    for key in foreign_keys:
        handler.add_action(ma.DropConstraintMigrationAction(
            table_name=new_table,
            constraint_name=key["name"]
        ), True)

    for constraint in constraints:
        removed.append(constraint["name"])
        handler.add_action(
            ma.DropConstraintMigrationAction(
                table_name=new_table,
                constraint_name=constraint["name"]
            )
        )
    indexes = inspector.get_indexes(table)
    for index in indexes:
        if index["name"] not in removed:
            handler.add_action(
                ma.DropIndexMigrationAction(
                    table_name=new_table,
                    index_name=index["name"]
                )
            )


def create_changelog_table(context: Context, new: Model, handler: MigrationHandler, rename: MigrateRename):
    table_name = get_pg_name(get_table_name(new, TableType.CHANGELOG))
    pkey_type = commands.get_primary_key_type(context, new.backend)
    handler.add_action(ma.CreateTableMigrationAction(
        table_name=table_name,
        columns=[
            sa.Column('_id', BIGINT, primary_key=True, autoincrement=True),
            sa.Column('_revision', sa.String),
            sa.Column('_txn', pkey_type, index=True),
            sa.Column('_rid', pkey_type),
            sa.Column('datetime', sa.DateTime),
            sa.Column('action', sa.String(8)),
            sa.Column('data', JSONB)
        ]
    ))


def handle_new_file_type(context: Context, backend: PostgreSQL, inspector: Inspector, prop: Property, pkey_type: Any,
                         handler: MigrationHandler) -> list:
    name = get_column_name(prop)
    nullable = not prop.dtype.required
    columns = []
    columns += [
        sa.Column(f'{name}._id', sa.String, nullable=nullable),
        sa.Column(f'{name}._content_type', sa.String, nullable=nullable),
        sa.Column(f'{name}._size', BIGINT, nullable=nullable)
    ]
    if BackendFeatures.FILE_BLOCKS in prop.dtype.backend.features:
        columns += [
            sa.Column(f'{name}._bsize', sa.Integer, nullable=nullable),
            sa.Column(f'{name}._blocks', ARRAY(pkey_type, ), nullable=nullable),
        ]
    new_table = get_pg_name(get_table_name(prop, TableType.FILE))
    if not inspector.has_table(new_table):
        handler.add_action(ma.CreateTableMigrationAction(
            table_name=new_table,
            columns=[
                sa.Column('_id', pkey_type, primary_key=True),
                sa.Column('_block', sa.LargeBinary)
            ]
        ))
    return columns


def handle_new_array_type(context: Context, backend: PostgreSQL, inspector: Inspector, prop: Property, pkey_type: Any,
                          handler: MigrationHandler):
    columns = []
    if isinstance(prop.dtype, Array) and prop.dtype.items:
        if prop.list is None:
            columns.append(sa.Column(prop.place, JSONB))

        if isinstance(prop.dtype.items.dtype, File):
            new_columns = handle_new_file_type(context, backend, inspector, prop.dtype.items, pkey_type, handler)
        elif isinstance(prop.dtype.items.dtype, Array):
            new_columns = handle_new_array_type(context, backend, inspector, prop.dtype.items, pkey_type, handler)
        elif isinstance(prop.dtype.items.dtype, Object):
            new_columns = handle_new_object_type(context, backend, inspector, prop.dtype.items, pkey_type, handler)
        else:
            new_columns = commands.prepare(context, backend, prop.dtype.items)
        if not isinstance(new_columns, list):
            new_columns = [new_columns]
        for column in new_columns:
            if not isinstance(column, sa.Column):
                new_columns.remove(column)

        new_table = get_pg_name(get_table_name(prop, TableType.LIST))
        if not inspector.has_table(new_table):
            main_table_name = get_pg_name(get_table_name(prop.model))
            handler.add_action(ma.CreateTableMigrationAction(
                table_name=new_table,
                columns=[
                    sa.Column('_txn', pkey_type, index=True),
                    sa.Column('_rid', pkey_type, sa.ForeignKey(
                        f'{main_table_name}._id', ondelete='CASCADE',
                    ), index=True),
                    *new_columns
                ]
            ))
    return columns


def handle_new_object_type(context: Context, backend: PostgreSQL, inspector: Inspector, prop: Property, pkey_type: Any,
                           handler: MigrationHandler):
    columns = []
    if isinstance(prop.dtype, Object) and prop.dtype.properties:
        for new_prop in prop.dtype.properties.values():
            if prop.name.startswith('_') and prop.name not in ('_revision',):
                continue
            if isinstance(new_prop.dtype, File):
                columns = handle_new_file_type(context, backend, inspector, new_prop, pkey_type, handler)
            elif isinstance(new_prop.dtype, Array):
                columns = handle_new_array_type(context, backend, inspector, new_prop, pkey_type, handler)
            elif isinstance(new_prop.dtype, Object):
                columns = handle_new_object_type(context, backend, inspector, new_prop, pkey_type, handler)
            else:
                columns = commands.prepare(context, backend, new_prop)

            if not isinstance(columns, list):
                columns = [columns]
            for column in columns:
                if not isinstance(column, sa.Column):
                    columns.remove(column)
    return columns


def get_prop_names(prop: Property):
    name = prop.name
    if isinstance(prop.dtype, Ref):
        if commands.identifiable(prop):
            name = f'{name}._id'
        else:
            for refprop in prop.dtype.refprops:
                yield f'{name}.{refprop.name}'
    yield name


def json_has_key(backend: PostgreSQL, column: sa.Column, table: sa.Table, key: str):
    with backend.engine.begin() as connection:
        query = sa.select(table.select().where(column.has_key(key)).exists())
        return connection.execute(query).scalar()


def jsonb_keys(backend: PostgreSQL, column: sa.Column, table: sa.Table):
    with backend.engine.begin() as connection:
        keys = sa.func.jsonb_object_keys(
            column
        )
        query = sa.select(
            [keys]
        ).select_from(table).group_by(keys)
        return [result[0] for result in connection.execute(query)]


def empty_table(backend: PostgreSQL, table: sa.Table) -> bool:
    with backend.engine.begin() as connection:
        # Fetching one row with limit is more efficient
        query = sa.select(1).select_from(table).limit(1)
        result = connection.execute(query)
        return result.fetchone() is None


def name_key(name: str):
    return name


def model_name_key(model: str) -> str:
    return get_pg_table_name(model)


def is_name_complex(name: str):
    return '.' in name or '@' in name


def is_prop_complex(prop: Property):
    return isinstance(prop.dtype, (Text, File, Ref))


def is_name_or_property_complex(name: str, prop: Property):
    is_complex = is_name_complex(name)
    if not is_complex:
        return is_prop_complex(prop)
    return is_complex


def is_name_or_column_complex(name: str, col: sa.Column):
    is_complex = is_name_complex(name)
    if not is_complex:
        return is_column_complex(col)
    return is_complex


def is_column_complex(col: sa.Column):
    return isinstance(col.type, (JSONB, JSON))


def property_and_column_name_key(
    item: Union[sa.Column, Property],
    rename,
    table: sa.Table,
    model: Model,
    root_name: str = ""
) -> str:
    # Mapping concept is to prioritize complex types over simple
    # new types take priority over old
    # Column is always old, Property is always new

    if isinstance(item, sa.Column):
        # Mapping order
        # Replace existing edge case -> Indirect renaming / removal edge case -> New name -> Old name

        name = item.name
        new_name = rename.get_column_name(table.name, name, True, root_value=root_name)
        full_name = rename.get_column_name(table.name, name)

        column_renamed = name_changed(name, new_name)
        column_directly_renamed = name_changed(name, full_name)

        # Check for edge case when you have old columns: column_one, column_two
        # new manifest only hase column_one, but
        # rename provides "column_two": "column_one"
        # meaning, you need to remove old "column_one" and rename old "column_two" to "column_one"
        if not column_renamed:
            old_name = rename.get_old_column_name(table.name, name)
            if name_changed(name, old_name):
                return name

        if column_renamed and not column_directly_renamed:
            new_prop = model.flatprops[new_name]
            if not is_name_or_property_complex(new_name, new_prop):
                return name

        return get_root_attr(new_name, initial_root=root_name)
    elif isinstance(item, Property):
        # Mapping order
        # New Property (complex) -> Old column (complex) -> New Property

        name = get_column_name(item)
        old_name = rename.get_old_column_name(table.name, name, True, root_value=root_name)
        old_full_name = rename.get_old_column_name(table.name, name)

        property_directly_renamed = name_changed(name, old_full_name)

        if is_name_or_property_complex(name, item):
            return get_root_attr(name, initial_root=root_name)

        if property_directly_renamed:
            if old_full_name in table.columns:
                col = table.columns[old_full_name]
                if is_column_complex(col):
                    return get_root_attr(old_full_name, initial_root=root_name)
            elif old_name in table.columns:
                col = table.columns[old_name]
                if is_column_complex(col):
                    return get_root_attr(old_name, initial_root=root_name)

        return get_root_attr(name, initial_root=root_name)


@dataclasses.dataclass
class MigratePostgresMeta:
    inspector: Inspector
    rename: MigrateRename
    handler: MigrationHandler


@dataclasses.dataclass
class JSONColumnMigrateMeta:
    column: sa.Column
    keys: List[str] = dataclasses.field(default_factory=list)
    new_keys: Dict[str, str] = dataclasses.field(default_factory=dict)
    full_remove: bool = dataclasses.field(default=True)
    cast_to: Tuple[sa.Column, str] = dataclasses.field(default=None)
    new_name: str = dataclasses.field(default=None)
    empty: bool = False

    def initialize(self, backend, table):
        self.keys = jsonb_keys(backend, self.column, table)
        self.empty = empty_table(backend, table)

    def add_new_key(self, old_key: str, new_key: str):
        if old_key in self.keys and old_key not in self.new_keys:
            self.new_keys[old_key] = new_key


@dataclasses.dataclass
class MigrateModelMeta:
    json_columns: Dict[str, JSONColumnMigrateMeta] = dataclasses.field(default_factory=dict)
    unique_constraint_states: Dict[str, bool] = dataclasses.field(default_factory=lambda: defaultdict(lambda: False))
    foreign_constraint_states: Dict[str, bool] = dataclasses.field(default_factory=lambda: defaultdict(lambda: False))
    index_states: Dict[str, bool] = dataclasses.field(default_factory=lambda: defaultdict(lambda: False))

    def initialize(
        self,
        backend: PostgreSQL,
        table: sa.Table,
        columns: List[sa.Column],
        inspector: Inspector
    ):
        for column in columns:
            # Add JSONB to meta (JSONB has different handle system)
            if isinstance(column.type, JSONB):
                self.__add_json_column(
                    backend,
                    table,
                    column
                )

        constraints = inspector.get_unique_constraints(table.name)
        for constraint in constraints:
            if not _reserved_constraint(constraint):
                self.unique_constraint_states[constraint['name']] = False

        constraints = inspector.get_foreign_keys(table.name)
        for constraint in constraints:
            self.foreign_constraint_states[constraint['name']] = False

        indexes = inspector.get_indexes(table.name)
        for index in indexes:
            if not _reserved_constraint(index):
                self.index_states[index['name']] = False

    def handle_unique_constraint(self, constraint: str):
        self.unique_constraint_states[constraint] = True
        self.index_states[constraint] = True

    def handle_foreign_constraint(self, constraint: str):
        self.foreign_constraint_states[constraint] = True

    def handle_index(self, constraint: str):
        self.index_states[constraint] = True

    def __add_json_column(self, backend: PostgreSQL, table: sa.Table, column: sa.Column):
        meta = JSONColumnMigrateMeta(
            column=column,
        )
        meta.initialize(backend, table)
        self.json_columns[column.name] = meta


def _reserved_constraint(constraint: dict) -> bool:
    return all(column_name.startswith('_') for column_name in constraint["column_names"])


def is_internal_ref(dtype: Ref):
    prop = dtype.prop
    return commands.identifiable(prop)


def handle_internal_ref_to_scalar_conversion(
    context: Context,
    backend: PostgreSQL,
    meta: MigratePostgresMeta,
    table: sa.Table,
    old_columns: List[sa.Column],
    new_property: Property,
    **kwargs
) -> bool:
    if isinstance(old_columns, sa.Column):
        old_columns = ensure_list(old_columns)

    if not old_columns or not new_property:
        return False

    # Skip ref 4 -> ref 3
    if isinstance(new_property.dtype, Ref):
        return False

    # Check if columns are from ref 4 (can only have 1 column)
    if not (len(old_columns) == 1 and isinstance(old_columns[0], sa.Column)):
        return False

    ref_col = old_columns[0]
    # Check if it is internal (should end with '._id'
    if not ref_col.name.endswith('._id'):
        return False

    inspector = meta.inspector
    handler = meta.handler
    rename = meta.rename

    manifest = new_property.model.manifest

    constraints = inspector.get_foreign_keys(table.name)
    ref_model = None
    table_name = None
    # Try to find referred table's matching model
    for constraint in constraints:
        if constraint['constrained_columns'] == [ref_col.name]:
            table_name = constraint['referred_table']
            if commands.has_model(context, manifest, table_name):
                ref_model = commands.get_model(context, manifest, table_name)
            else:
                # In case table name has been truncated, need to loop through all models and convert their names to pg
                all_models = commands.get_models(context, manifest)
                for model in all_models.values():
                    if get_pg_name(model.name) == table_name:
                        ref_model = model
                        break
            break

    if not ref_model:
        return False

    ref_primary_keys = get_spinta_primary_keys(
        table_name=table_name,
        model=ref_model,
        inspector=inspector
    )

    if len(ref_primary_keys) > 1:
        raise MigrateScalarToRefTooManyKeys(new_property.dtype, primary_keys=[key for key in ref_primary_keys])

    ref_primary_property = ref_model.flatprops[ref_primary_keys[0]]
    ref_primary_column = commands.prepare(context, backend, ref_primary_property)
    column_name = get_pg_name(get_column_name(new_property))
    updated_kwargs = adjust_kwargs(kwargs, {
        'foreign_key': True
    })

    commands.migrate(context, backend, meta, table, NA, new_property, **updated_kwargs)
    table_name = get_pg_table_name(rename.get_table_name(table.name))
    foreign_table_name = get_pg_table_name(get_table_name(ref_model))
    handler.add_action(
        ma.DowngradeTransferDataMigrationAction(
            table_name,
            foreign_table_name,
            ref_col,
            {
                column_name: ref_primary_column
            },
            '_id'
        ),
        foreign_key=True
    )
    commands.migrate(context, backend, meta, table, ref_col, NA, **updated_kwargs)
    return True


def extract_target_column(rename: MigrateRename, columns: list, table: sa.Table, prop: Property):
    full_name = rename.get_old_column_name(table.name, prop.name)
    if isinstance(columns, list):
        for col in columns:
            if isinstance(col, sa.Column) and col.name == full_name:
                return [col]
    return columns


def adjust_kwargs(kwargs: dict, new: dict) -> dict:
    copied = kwargs.copy()
    copied.update(new)
    return copied


def extract_literal_name_from_column(
    column: sa.Column,
) -> str:
    type_ = column.type.compile(dialect=postgresql.dialect())

    # Convert sa.Float, to postgresql DOUBLE PRECISION type
    if isinstance(column.type, sa.Float):
        type_ = 'DOUBLE PRECISION'

    return type_


def extract_using_from_columns(
    old_column: sa.Column,
    new_column: sa.Column,
    type_
):
    using = None
    if (isinstance(old_column.type, geoalchemy2.types.Geometry)
        and isinstance(new_column.type, geoalchemy2.types.Geometry)
        and old_column.type.srid != new_column.type.srid
    ):
        srid_name = old_column
        srid = new_column.type.srid
        if old_column.type.srid == -1:
            srid_name = sa.func.ST_SetSRID(old_column, 4326)
        if new_column.type.srid == -1:
            srid = 4326
        using = sa.func.ST_Transform(srid_name, srid).compile(compile_kwargs={"literal_binds": True})
    elif type_ is not None:
        using = sa.func.cast(old_column, type_).compile(compile_kwargs={"literal_binds": True})

    return using


# Match [
#   (
#       (old_column_name, old_type),
#       (new_column_name, new_type)
#   )
# ]
def generate_type_missmatch_exception_details(
    columns: list
):
    result = ''
    for pair in columns:
        old_data = pair[0]
        new_data = pair[1]
        result += f'\t\'{old_data[0]}\' [{old_data[1]}] -> \'{new_data[0]}\' [{new_data[1]}]\t'
        if old_data[1] == new_data[1]:
            result += f'\'{old_data[1]}\' == \'{new_data[1]}\'\n'
        else:
            result += f'\'{old_data[1]}\' != \'{new_data[1]}\'\t<= Incorrect\n'
    return result


def constraint_with_name(constraints: list, constraint_name: str):
    try:
        return next(constraint for constraint in constraints if constraint["name"] == constraint_name)
    except StopIteration:
        return None


def constraint_with_columns(constraints: list, column_names: list[str]):
    try:
        return next(constraint for constraint in constraints if constraint["column_names"] == column_names)
    except StopIteration:
        return None


def index_with_columns(indexes: list, column_names: list[str], condition: Callable[[dict], bool] = lambda index: True):
    try:
        return next(index for index in indexes if index["column_names"] == column_names and condition(index))
    except StopIteration:
        return None


def index_with_name(indexes: list, index_name: str, condition: Callable[[dict], bool] = lambda index: True):
    try:
        return next(index for index in indexes if index["name"] == index_name and condition(index))
    except StopIteration:
        return None


def index_not_handled_condition(meta: MigrateModelMeta):
    return lambda index: not meta.index_states[index['name']]


def contains_unique_constraint(constraints: list, column_name: str):
    return any(constraint["column_names"] == [column_name] for constraint in constraints)


def contains_constraint_name(constraints: list, constraint_name: str):
    return any(constraint["name"] == constraint_name for constraint in constraints)


def contains_index(indexes: list, column_name: str):
    return any(index["column_names"] == [column_name] for index in indexes)


def contains_foreign_key_with_table_columns(constraints: list, table_name: str, column_names: list[str]):
    return any(
        constraint["constrained_columns"] == column_names and constraint["referred_table"] == table_name for constraint
        in constraints)


def handle_unique_constraint_migration(
    table: sa.Table,
    table_name: str,
    old: sa.Column,
    new: sa.Column,
    column_name: str,
    handler: MigrationHandler,
    inspector: Inspector,
    foreign_key: bool,
    renamed: bool,
    meta: MigrateModelMeta
):
    if not new.unique:
        return

    unique_name = get_pg_constraint_name(table_name, column_name)

    if meta.unique_constraint_states[unique_name]:
        return

    unique_constraints = inspector.get_unique_constraints(table_name=table.name)
    constraint_column = old.name if renamed else column_name

    meta.handle_unique_constraint(unique_name)
    old_constraint = constraint_with_columns(unique_constraints, [constraint_column])
    if old_constraint and old_constraint['name'] == unique_name:
        return

    if not contains_constraint_name(unique_constraints, unique_name):
        if old_constraint:
            meta.handle_unique_constraint(old_constraint['name'])
            handler.add_action(ma.RenameConstraintMigrationAction(
                table_name=table_name,
                old_constraint_name=old_constraint['name'],
                new_constraint_name=unique_name
            ))
            return

        handler.add_action(ma.CreateUniqueConstraintMigrationAction(
            constraint_name=unique_name,
            table_name=table_name,
            columns=[column_name]
        ), foreign_key)
        return

    if not contains_unique_constraint(unique_constraints, constraint_column):
        handler.add_action(ma.DropConstraintMigrationAction(
            constraint_name=unique_name,
            table_name=table_name,
        ), foreign_key)

        handler.add_action(ma.CreateUniqueConstraintMigrationAction(
            constraint_name=unique_name,
            table_name=table_name,
            columns=[column_name]
        ), foreign_key)


def _index_using_suffix(
    column: sa.Column
):
    if isinstance(column.type, geoalchemy2.types.Geometry):
        return "GIST"

    return None


def _requires_index(
    column: sa.Column,
    skip_unique: bool = True
) -> bool:
    if isinstance(column.type, geoalchemy2.types.Geometry):
        return True

    # This is handled with UniqueConstraint handler
    if skip_unique and column.unique:
        return False

    return column.index


def handle_index_migration(
    table: sa.Table,
    table_name: str,
    old: sa.Column,
    new: sa.Column,
    column_name: str,
    handler: MigrationHandler,
    inspector: Inspector,
    foreign_key: bool,
    renamed: bool,
    meta: MigrateModelMeta
):
    if not _requires_index(new):
        return

    index_name = get_pg_index_name(table_name=table_name, columns=[column_name])
    if meta.index_states[index_name]:
        return

    constraint_column = old.name if renamed else column_name
    indexes = inspector.get_indexes(table_name=table.name)
    using = _index_using_suffix(new)

    # Check unhandled index with same columns
    existing_index = index_with_columns(indexes, [constraint_column], condition=index_not_handled_condition(meta))
    meta.handle_index(index_name)
    if existing_index is not None:
        if existing_index['name'] == index_name:
            return

        meta.handle_index(existing_index["name"])
        handler.add_action(ma.RenameIndexMigrationAction(
            old_index_name=existing_index["name"],
            new_index_name=index_name
        ))
        return

    # Check index with existing name
    if contains_index(indexes, index_name):
        existing_index = index_with_name(indexes, index_name)
        if existing_index['column_names'] == [constraint_column]:
            return

        handler.add_action(
            ma.DropIndexMigrationAction(
                index_name=index_name,
                table_name=table_name
            ), foreign_key)
        handler.add_action(
            ma.CreateIndexMigrationAction(
                index_name=index_name,
                table_name=table_name,
                columns=[constraint_column],
                using=using
            ), foreign_key)
        return

    handler.add_action(
        ma.CreateIndexMigrationAction(
            index_name=index_name,
            table_name=table_name,
            columns=[constraint_column],
            using=using
        ), foreign_key)


def extract_sqlalchemy_columns(data: list) -> List[sa.Column]:
    return [item for item in data if isinstance(item, sa.Column)]


def reduce_columns(data: list) -> Union[sa.Column, list[sa.Column]]:
    return data[0] if len(data) == 1 else data


def is_internal(
    columns: List[sa.Column],
    base_name: str,
    table_name: str,
    ref_table_name: str,
    inspector: Inspector
) -> bool:
    column_name = get_pg_column_name(f'{base_name}._id')
    contains_column = any(column.name == column_name for column in columns)

    if not contains_column:
        return False

    foreign_keys = inspector.get_foreign_keys(table_name)
    return contains_foreign_key_with_table_columns(foreign_keys, ref_table_name, [column_name])


def split_columns(
    columns: List[sa.Column],
    base_name: str,
    target_base_name: str,
    internal: bool,
    ref_table_primary_key_names: List[str],
    target_primary_column_names: List[str],
    target_children_column_names: List[str]
):
    all_column_names = [column.name for column in columns]
    primary_columns = []
    children_columns = []

    # If we know that old columns contain internal ref key, that means we can extract it and everything else are children
    if internal:
        for column in columns:
            column_name = column.name
            if column_name.endswith('_id') and column_name.replace(f'{base_name}.', '') == '_id':
                primary_columns = [column]
                continue

            children_columns.append(column)
        return primary_columns, children_columns

    # Check if all old columns contain all target children columns
    all_children_match = all(column_name.replace(target_base_name, base_name) in all_column_names for column_name in
                             target_children_column_names)

    # Check if all old columns contain all target primary key columns
    all_primary_match = all(column_name.replace(target_base_name, base_name) in all_column_names for column_name in
                            target_primary_column_names)

    if all_primary_match:
        # Primary keys are priority so if they all match we assume:
        # primary_columns = all target_primary_columns
        # children_columns = everything else
        for column in columns:
            column_name = column.name.replace(base_name, target_base_name)
            if column_name in target_primary_column_names:
                primary_columns.append(column)
                continue

            children_columns.append(column)
        return primary_columns, children_columns

    elif all_children_match:
        # If all primary keys do not match, but children do, we need to figure out what is the foreign key
        remaining_column_names = [column.name for column in columns if
                                  column.name.replace(base_name, target_base_name) not in target_children_column_names]
        all_remaining_match = all(
            f'{base_name}.{column_name}' in remaining_column_names for column_name in ref_table_primary_key_names)

        if all_remaining_match:
            # If all renaming columns, after children, match ref tables primary keys, we assume that foreign keys
            # are ref table primary keys
            # primary_columns = all ref primary keys
            # children_columns = everything else
            for column in columns:
                column_name = column.name.replace(f'{base_name}.', '')
                if column_name in ref_table_primary_key_names:
                    primary_columns.append(column)
                    continue

                children_columns.append(column)
            return primary_columns, children_columns

        # We could not find primary key source, so we assume that everything that remains is primary key
        for column in columns:
            column_name = column.name
            if column_name in remaining_column_names:
                primary_columns.append(column)
                continue

            children_columns.append(column)
        return primary_columns, children_columns

    else:
        # Nothing matches we can only try to guess
        converted_ref_names = [f'{base_name}.{column_name}' for column_name in ref_table_primary_key_names]
        all_ref_primary_keys_match = all(
            column_name in all_column_names for column_name in converted_ref_names)

        if all_ref_primary_keys_match:
            # Since all ref model's primary keys are in the list, we can assume, that this is the foreign key
            for column in columns:
                column_name = column.name
                if column_name in converted_ref_names:
                    primary_columns.append(column)
                    continue

                children_columns.append(column)
            return primary_columns, children_columns

    # If nothing worked then we assume that all columns are foreign keys
    for column in columns:
        primary_columns.append(column.name)
    return primary_columns, children_columns


def get_spinta_primary_keys(
    table_name: str,
    model: Model,
    inspector: Inspector
) -> List[str]:
    """Extracts `manifest` declared primary keys (from internal PostgresSql)

    Args:
        table_name: old table's name
        model: new table's model
        inspector: SQLAlchemy Inspector object

    Since spinta on internal backend does not actually store primary keys as `PrimaryKey`
    You can only know if primary key was set in manifest through `UniqueConstraint`
    `PrimaryKey` is reserved for `_id` property
    Issue is that it is not mandatory to set primary key
    Also you can explicitly create unique constraints through manifest, meaning you cannot be certain
    that the `UniqueConstraint` you find is actually primary key
    """

    unique_constraints = inspector.get_unique_constraints(table_name)
    unique_constraint_columns = [constraint['column_names'] for constraint in unique_constraints]

    if not unique_constraint_columns:
        return []

    if not model.external.unknown_primary_key:
        # If model contains primary key, we might be able to find column combination
        # which would take priority
        primary_property_names = [prop.place for prop in model.external.pkeys]
        for constraint in unique_constraints:
            if set(constraint) == set(primary_property_names):
                return constraint

    # New model does have declared primary key, making it hard to predict if table had it set before
    if len(unique_constraint_columns) == 1:
        # If there is only 1 combination
        return unique_constraint_columns[0]
    return []


def remap_and_rename_columns(
    base_name: str,
    columns: List[sa.Column],
    table_name: str,
    ref_table_name: str,
    rename: MigrateRename
) -> dict:
    result = {}
    for column in columns:
        name = rename.get_column_name(table_name, column.name)
        # Handle nested renaming from 2 tables
        if column.name.startswith(base_name):
            leaf_name = column.name.removeprefix(base_name)
            if leaf_name.startswith('.'):
                leaf_name = leaf_name[1:]

            base_renamed = nested_column_rename(base_name, table_name, rename)
            leaf_renamed = nested_column_rename(leaf_name, ref_table_name, rename)
            name = f"{base_renamed}.{leaf_renamed}"
        result[name] = column
    return result


def remove_property_prefix_from_column_name(
    column_name: str,
    prop: Property,
) -> str:
    return column_name.replace(f'{prop.place}.', '', 1)


def zip_and_migrate_properties(
    context: Context,
    backend: PostgreSQL,
    old_table: sa.Table,
    new_model: Model,
    old_columns: List[sa.Column],
    new_properties: List[Property],
    meta: MigratePostgresMeta,
    rename: MigrateRename,
    root_name: str = "",
    **kwargs
):
    zipped_items = zipitems(
        old_columns,
        new_properties,
        lambda x: property_and_column_name_key(x, rename, old_table, new_model, root_name=root_name)
    )
    for zipped_item in zipped_items:
        old_columns = []
        new_properties = []
        for old_column, new_property in zipped_item:
            # Ignore deleted / reserved properties
            if new_property and new_property.name.startswith('_'):
                continue

            if old_column not in old_columns:
                old_columns.append(old_column)
            if new_property and new_property not in new_properties:
                new_properties.append(new_property)

        if len(old_columns) == 1:
            old_columns = old_columns[0]
        elif not old_columns:
            old_columns = NA
            # If neither column nor property is matched skip it
            if not new_properties:
                continue
        if new_properties:
            for new_property in new_properties:
                handled = handle_internal_ref_to_scalar_conversion(
                    context,
                    backend,
                    meta,
                    old_table,
                    old_columns,
                    new_property,
                    **kwargs
                )

                if not handled:
                    commands.migrate(context, backend, meta, old_table, old_columns, new_property,
                                     **kwargs)
        else:
            commands.migrate(context, backend, meta, old_table, old_columns, NA,
                             **kwargs)
